from .reference import Reference
from enum import Enum
import importlib.resources
import numpy
import pandas


class BOWERMANN_2022(Reference):
    """
    Implementation of the race-neutral reference equations for FEV1, FVC and FEV1/FVC.
    
    Citation 1:
    Bowerman C, Bhakta NR, Brazzale D, Cooper BR, Cooper J, Gochicoa-Rangel L, Haynes J, Kaminsky DA, Lan LTT, 
    Masekela R, McCormack MC, Steenbruggen I, Stanojevic S. A Race-neutral Approach to the Interpretation of 
    Lung Function Measurements. Am J Respir Crit Care Med. 2023 Mar 15;207(6):768-774. 
    doi: 10.1164/rccm.202205-0963OC. PMID: 36383197.
    """

    class Parameters(Enum):
        FEV1 = 1
        FVC = 2
        FEV1FVC = 3

    def __init__(self):
        """
        No height range given.
        """
        self.__lookup, self.__splines = self.__load_lookup_table()

    def __load_lookup_table(self) -> tuple:
        """
        Loads and stores the coefficient and splines values.
        :return: both files as pandas dataframe
        """
        lookup_path = importlib.resources.open_binary('pyspiro.data', 'bowermann_2022_splines.csv')
        splines_path = importlib.resources.open_binary('pyspiro.data', 'bowermann_2022_coefficients.csv')
        lookup = pandas.read_csv(lookup_path, delimiter=";").set_index("age")
        splines = pandas.read_csv(splines_path, delimiter=";").set_index("var")
        self._age_range: tuple = (min(lookup.index), max(lookup.index))
        return lookup, splines

    def __get_splines(self, sex: int, age: float, parameter: int):
        """
        Yields the appropriated splines values based on the parameter, sex and age.
        :param sex: self.Sex enumration as integer value
        :param age: age as float
        :param parameter: self.Parameter enumeration as integer value
        """
        for i in ("Sspline", "Mspline", "Lspline"):
            yield self.__lookup["%s_%ss_%s" % (self.Parameters(parameter).name, self.Sex(sex).name.lower(), i)].loc[age] # Change sth here

    def lms(self, sex: int, age: float, height: float, parameter: int, value: float) -> tuple:
        """
        Calculate l, m and s values for the given parameters.
        """
        age = self.validate_range(round(age * 4) / 4, self._age_range, "age")
        if age is pandas.NA:
            return pandas.NA, pandas.NA, pandas.NA
        sspline, mspline, lspline = self.__get_splines(sex, age, parameter) #LSpline not used.
        c = self.__splines["%s_%ss" % (self.Parameters(parameter).name, self.Sex(sex).name.lower())]

        if self.Parameters(parameter).name in ['FEV1', 'FVC']:
            l = c.loc["q0"]
        elif self.Parameters(parameter).name in ['FEV1FVC']:
            l = c.loc["q0"] + (c.loc["q1"] * numpy.log(age))

        m = numpy.exp(c.loc["a0"] + (c.loc["a1"] * numpy.log(height)) + (c.loc["a2"] * numpy.log(age)) + mspline)
        s = numpy.exp(c.loc["p0"] + (c.loc["p1"] * numpy.log(age)) + sspline)

        return l, m, s

    def percent(self, sex: int, age: float, height: float, parameter: int, value: float, silent = True):
        """
        Returns % of predicted value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else round(( value / m ) * 100, 2)

    def zscore(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns z-score value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else (((value/m)**l) - 1) / (l * s)

    def lln(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns lower limit of normal, by convention the lower 5th percentile.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))

    def all(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns all values at once (percent, z-score and lln).
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        if (l is pandas.NA or m is pandas.NA or s is pandas.NA):
            return pandas.NA
        else:
            return round(( value / m ) * 100, 2), (((value/m)**l) - 1) / (l * s), numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))


from .reference import Reference
from enum import Enum
import importlib.resources
import numpy
import pandas


class SCHULZ_2013(Reference):
    """
    A dataset for impulse oscillometry reference equations published by Schulz et al. 2013.

    citation:
    Schulz H, Flexeder C, Behr J, Heier M, Holle R, Huber RM, Jörres RA, Nowak D, Peters A, 
    Wichmann HE, Heinrich J, Karrasch S; KORA Study Group. Reference values of impulse oscillometric 
    lung function indices in adults of advanced age. PLoS One. 2013 May 15;8(5):e63366. 
    doi: 10.1371/journal.pone.0063366. PMID: 23691036; PMCID: PMC3655177.
    """


    class Parameters(Enum):
        R10 = 1
        R15 = 2
        R25 = 3
        R35 = 4
        X10 = 5
        X20 = 6
        X25 = 7
        X35 = 8

    def __init__(self):
        """
        No height range given.
        """
        self.__lookup = self.__load_lookup_table()

    def __load_lookup_table(self) -> tuple:
        """
        Loads and stores the coefficient and splines values.
        :return: both files as pandas dataframe
        """
        splines_path = importlib.resources.open_binary('pyspiro.data', 'schulz_2013_splines.csv')
        splines = pandas.read_csv(splines_path, delimiter=";").set_index("Var")
        return splines

    def __get_splines(self, sex: int, pct: float, parameter: int):
        """
        Yields the appropriated splines values based on the parameter, sex and age.
        :param sex: self.Sex enumration as integer value
        :param pct: percentile as float, can be 0.05, 0.5 or 0.95
        :param parameter: self.Parameter enumeration as integer value
        """

        for i in ("intercept", "age", "height", "weight"):
            yield self.__lookup["%s_%ss" % (self.Parameters(parameter).name, self.Sex(sex).name.lower())].loc["%s_%s" % (i, pct)]

    def percentiles(self, sex: int, age: float, height: float, weight: float, parameter: int) -> tuple:
        """
        Return the 5, 50 and 95 percentile values.
        """
        if age is pandas.NA or sex is pandas.NA or height is pandas.NA or weight is pandas.NA:
            return pandas.NA, pandas.NA, pandas.NA

        q1, q2, q3, q4 = self.__get_splines(sex, pct = 0.05, parameter = parameter)
        q4 = 0 if pandas.isna(q4) else q4
        p5 = q1 + (q2 * age) + (q3 * height) + (q4 *weight)

        q1, q2, q3, q4 = self.__get_splines(sex, parameter = parameter, pct = 0.50)
        q4 = 0 if pandas.isna(q4) else q4
        p50 = q1 + (q2 * age) + (q3 * height) + (q4 *weight)

        q1, q2, q3, q4 = self.__get_splines(sex, parameter = parameter, pct = 0.95)
        q4 = 0 if pandas.isna(q4) else q4
        p95 = q1 + (q2 * age) + (q3 * height) + (q4 *weight)

        return p5, p50, p95

    def lms(self, sex: int, age: float, height: float, parameter: int, value: float) -> tuple:
        """
        Not implemented.
        """
        pass

    def percent(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Not implemented.        
        """
        pass

    def zscore(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Not implemented.
        """
        pass

    def lln(self, sex: int, age: float, height: float, weight: float, parameter: int, value: float):
        """
        Returns lower limit of normal, by convention the lower 5th percentile.
        """
        p5, p50, p95 = self.percentiles(sex, age, height, weight, parameter, value)
        return pandas.NA if (p5 is pandas.NA or p50 is pandas.NA or p95 is pandas.NA) else p5

    def all(self, sex: int, age: float, height: float, weight: float, parameter: int, value: float):
        """
        Returns all values at once (5th, 50th and 95th percentile).
        """
        p5, p50, p95 = self.percentiles(sex, age, height, weight, parameter, value)
        if (p5 is pandas.NA or p50 is pandas.NA or p95 is pandas.NA):
            return pandas.NA, pandas.NA, pandas.NA
        else:
            return p5, p50, p95


from .reference import Reference
from enum import Enum
import importlib.resources
import numpy
import pandas


class GLI_2021(Reference):
    """
    This global lung function was published in 2021 by Hall and colleagues and includes as variable sex, 
    age, height and ethnicity, providing functions for TLCO, DLCO, KCO and VA.

    It includes reference values for Caucasians aged 5-80 years.

    Hall GL, Filipow N, Ruppel G, Okitika T, Thompson B, Kirkby J, Steenbruggen I, Cooper BG, Stanojevic S; contributing 
    GLI Network members. Official ERS technical standard: Global Lung Function Initiative reference values for static lung 
    volumes in individuals of European ancestry. Eur Respir J. 2021 Mar 11;57(3):2000289. doi: 10.1183/13993003.00289-2020. 
    PMID: 33707167.
    """


    class Parameters(Enum):
        FRC = 1
        TLC = 2
        RV = 3
        RV_TLC = 4
        ERV = 5
        IC = 6
        VC = 7

    def __init__(self):
        """
        No height range given.
        """
        self.__lookup, self.__splines = self.__load_lookup_table()

    def __load_lookup_table(self) -> tuple:
        """
        Loads and stores the coefficient and splines values.
        :return: both files as pandas dataframe
        """
        lookup_path = importlib.resources.open_binary('pyspiro.data', 'gli_2021_splines.csv')
        splines_path = importlib.resources.open_binary('pyspiro.data', 'gli_2021_coefficients.csv')
        lookup = pandas.read_csv(lookup_path, delimiter=";").set_index("age")
        splines = pandas.read_csv(splines_path, delimiter=";").set_index("var")
        self._age_range: tuple = (min(lookup.index), max(lookup.index))
        return lookup, splines

    def __get_splines(self, sex: int, age: float, parameter: int):
        """
        Yields the appropriated splines values based on the parameter, sex and age.
        :param sex: self.Sex enumration as integer value
        :param age: age as float
        :param parameter: self.Parameter enumeration as integer value
        """
        for i in ("Sspline", "Mspline", "Lspline"):
            yield self.__lookup["%s_%ss_%s" % (self.Parameters(parameter).name, self.Sex(sex).name.lower(), i)].loc[age] # Change sth here

    def lms(self, sex: int, age: float, height: float, parameter: int, value: float) -> tuple:
        """
        Calculate l, m and s values for the given parameters.
        """
        age = self.validate_range(round(age * 4) / 4, self._age_range, "age")
        if age is pandas.NA:
            return pandas.NA, pandas.NA, pandas.NA
        sspline, mspline, lspline = self.__get_splines(sex, age, parameter) #LSpline not used.
        c = self.__splines["%s_%ss" % (self.Parameters(parameter).name, self.Sex(sex).name.lower())]
        
        if self.Parameters(parameter).name in ['FRC', 'TLC', 'RV', 'RV_TLC']:
            s = numpy.exp(float(c.loc["p0"]) + (float(c.loc["p1"]) * numpy.log(age)) + sspline)
        elif self.Parameters(parameter).name in ['ERV', 'IC', 'VC']:
            s = numpy.exp(float(c.loc["p0"]) + (float(c.loc["p1"]) * numpy.log(age)))

        m = numpy.exp(float(c.loc["a0"]) + (float(c.loc["a1"]) * numpy.log(height)) + (float(c.loc["a2"]) * numpy.log(age)) + mspline)
        l = float(c.loc["q0"])

        return l, m, s

    def percent(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns % of predicted value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else round(( value / m ) * 100, 2)

    def zscore(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns z-score value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else (((value/m)**l) - 1) / (l * s)

    def lln(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns lower limit of normal, by convention the lower 5th percentile.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))

    def all(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns all values at once (percent, z-score and lln).
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        if (l is pandas.NA or m is pandas.NA or s is pandas.NA):
            return pandas.NA
        else:
            return round(( value / m ) * 100, 2), (((value/m)**l) - 1) / (l * s), numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))


from .reference import Reference
from enum import Enum
import numpy as np
import pandas as pd


class KUSTER_2008(Reference):
    """
    A 2008 ERJ publication establishing reference values for lung function screening in ~8500 patients from the swiss LuftiBus trial.

    The ranges of application for these reference equations are ages 18-80 yrs and heights of 140-200 and 130-190 cm in males and 
    females, respectively.

    Kuster SP, Kuster D, Schindler C, Rochat MK, Braun J, Held L, Brändli O. Reference equations for lung 
    function screening of healthy never-smoking adults aged 18-80 years. Eur Respir J. 2008 Apr;31(4):860-8. 
    doi: 10.1183/09031936.00091407. Epub 2007 Dec 5. PMID: 18057057.
    """

    class Parameters(Enum):
        FVC = 1
        FVC_LLN = 2
        FEV1 = 3
        FEV1_LLN = 4
        MEF75 = 5
        MEF75_LLN = 6
        MEF50 = 7
        MEF50_LLN = 8
        MEF25 = 9
        MEF25_LLN = 10
        FEV1_FVC_P = 11
        FEV1_FVC_LLN = 12
        PEF = 13
        PEF_LLN = 14

    def __init__(self):
        self._age_range = (18, 80)
        self._height_female_range = (130, 190)
        self._height_male_range = (140, 200)

    def lms(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float) -> tuple:
        pass

    def zscore(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        pass

    def _check_conditions(self, sex: int, age: float, height: float):
        age = self.validate_range(round(age * 4) / 4, self._age_range, "age")
        height_range = self._height_female_range if sex == self.Sex["FEMALE"].value else self._height_male_range
        height = self.validate_range(height, height_range, "height")
        sex = self.check_tuple(sex, (0,1), "sex")
        return age, height, sex

    def percent(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        """
        In this case, only calculate the m (predicted value).
        """
        age, height, sex = self._check_conditions(sex, age, height)
        if age is pd.NA or height is pd.NA or sex is pd.NA:
            return pd.NA

        #print(f"[DEBUG] parameter={parameter} sex={sex} age={age} height={height}")
        if sex == self.Sex["FEMALE"].value:

            match parameter:
                case self.Parameters.FVC:
                    # FVC L = exp(-9.069+2.013 ln(H)+0.00847 A-0.000155A2)
                    m = np.exp( -9.069 + 2.013 * np.log(height) + 0.00847 * (age) - 0.000155 * (age ** 2) )
                case self.Parameters.FEV1:
                    # FEV1 L = exp(-8.397+1.865 ln(H)+0.00570 A-0.000150A2)
                    m = np.exp( -8.397 + 1.865 * np.log(height) + 0.00570 * (age) - 0.000150 * (age ** 2) )
                case self.Parameters.MEF75:
                    # MEF75 L·s−1 = exp(-2.716+0.867 ln(H)+0.00963 A-0.000140A2)
                    m = np.exp( -2.716 + 0.867 * np.log(height) + 0.00963 * (age) - 0.000140 * (age ** 2) )
                case self.Parameters.MEF50:
                    # MEF50 L·s−1 = exp(-2.131+0.674 ln(H)+0.00895 A-0.000180A2)
                    m = np.exp( -2.131 + 0.674 * np.log(height) + 0.00895 * (age) - 0.000180 * (age ** 2) )
                case self.Parameters.MEF25:
                    # MEF25 L·s−1 = exp(-4.861+1.145 ln(H)-0.01120 A-0.000096A2)
                    m = np.exp( -4.861 + 1.145 * np.log(height) - 0.01120 * (age) - 0.000096 * (age ** 2) )
                case self.Parameters.FEV1_FVC_P:
                    # FEV1/FVC % =  exp(+5.637-0.219 ln(H)-0.00249 A+0.000004A2)
                    m = np.exp( 5.637 - 0.219 * np.log(height) - 0.00249 * (age) + 0.000004 * (age ** 2) )
                case self.Parameters.PEF:
                    # PEF L·s−1 = exp(-4.794+1.316 ln(H)+0.00926 A-0.000143A2)
                    m = np.exp( -4.794 + 1.316 * np.log(height) + 0.00926 * (age) - 0.000143 * (age ** 2) )
                case _:
                    raise ValueError(f"Unknown parameter for percent calculation: {parameter}")
            return round(( value / m ) * 100, 2)

        elif sex == self.Sex["MALE"].value:

            match parameter:

                case self.Parameters.FVC:
                    # FVC L = exp(-10.258+2.280 ln(H)+0.00676A-0.000124A2)
                    m = np.exp( -10.258 + 2.280 * np.log(height) + 0.00676 * (age) - 0.000124 * (age ** 2) )
                case self.Parameters.FEV1:
                    # FEV1 L = exp(-8.957+2.014 ln(H)+0.00281A-0.000105A2)
                    m = np.exp( -8.957 + 2.014 * np.log(height) + 0.00281 * (age) - 0.000105 * (age ** 2) )
                case self.Parameters.MEF75:
                    # MEF75 L·s−1 = exp(-2.227+0.812 ln(H)+0.00977A-0.000132A2)
                    m = np.exp( -2.227 + 0.812 * np.log(height) + 0.00977 * (age) - 0.000132 * (age ** 2) )
                case self.Parameters.MEF50:
                    # MEF50 L·s−1 = exp(-3.055+0.911 ln(H)+0.00249A-0.000109A2)
                    m = np.exp( -3.055 + 0.911 * np.log(height) - 0.00249 * (age) - 0.000109 * (age ** 2) )
                case self.Parameters.MEF25:
                    # MEF25 L·s−1 = exp(-3.970+1.009 ln(H)-0.01645A-0.000020A2)
                    m = np.exp( -3.970 + 1.009 * np.log(height) - 0.01645 * (age) -0.000020 * (age ** 2) )
                case self.Parameters.FEV1_FVC_P:
                    # FEV1/FVC % = exp(+6.291-0.341 ln(H)-0.00441A+0.000026A2)
                    m = np.exp( 6.291 - 0.341 * np.log(height) - 0.00441 * (age) + 0.000026 * (age ** 2) ) 
                case self.Parameters.PEF:
                    # PEF L·s−1 = exp(-3.760+1.170 ln(H)+0.00706A-0.000110A2)
                    m = np.exp( -3.760 + 1.170 * np.log(height) + 0.00706 * (age) - 0.000110 * (age ** 2) )
                case _:
                    raise ValueError(f"Unknown parameter for percent calculation: {parameter}")
            return round(( value / m ) * 100, 2)


    def lln(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        age, height, sex = self._check_conditions(sex, age, height)
        if age is pd.NA or height is pd.NA or sex is pd.NA:
            return pd.NA

        if sex == self.Sex["FEMALE"].value:

            match parameter:
                case self.Parameters.FVC_LLN:
                    # FVC L = exp(-9.213 + 2.013 ln(H) + 0.00616A - 0.000155A^2)
                    lln = np.exp( -9.213 + 2.013 * np.log(height) + 0.00616 * (age) - 0.000155 * (age ** 2) )
                case self.Parameters.FEV1_LLN:
                    # FEV1 L = exp(-8.521 + 1.865 ln(H) + 0.00357A - 0.000150A^2)
                    lln = np.exp( -8.521 + 1.865 * np.log(height) + 0.00357 * (age) - 0.000150 * (age ** 2) )
                case self.Parameters.MEF75_LLN:
                    # MEF75 L·s−1 = exp(-2.977 + 0.867 ln(H) + 0.00698A - 0.000140A^2)
                    lln = np.exp( -2.977 + 0.867 * np.log(height) + 0.00698 * (age) - 0.000140 * (age ** 2) )
                case self.Parameters.MEF50_LLN:
                    # MEF50 L·s−1 = exp(-2.374 + 0.674 ln(H) + 0.00330A - 0.000180A^2)
                    lln = np.exp( -2.374 + 0.674 * np.log(height) + 0.00330 * (age) - 0.000180 * (age ** 2) )
                case self.Parameters.MEF25_LLN:
                    # MEF25 L·s−1 = exp(-5.140 + 1.145 ln(H) - 0.02002A - 0.000096A^2)
                    lln = np.exp( -5.140 + 1.145 * np.log(height) - 0.02002 * (age) - 0.000096 * (age ** 2) )
                case self.Parameters.FEV1_FVC_LLN:
                    # FEV1/FVC % =	exp(+5.524-0.219 ln(H)-0.00313A+0.000004A2)
                    lln = np.exp( 5.524 - 0.219 * np.log(height) - 0.00313 * (age) + 0.000004 * (age ** 2) )
                case self.Parameters.PEF_LLN:
                    # PEF L·s−1 = exp(-5.032+1.316 ln(H)+0.00767A-0.000143A2)
                    lln = np.exp( -5.032 + 1.316 * np.log(height) + 0.00767 * (age) - 0.000143 * (age ** 2) )
                case _:
                    raise ValueError(f"Unknown parameter for percent calculation: {parameter}")
            return lln

        elif sex == self.Sex["MALE"].value:

            match parameter:
                case self.Parameters.FVC_LLN:
                    # FVC L = exp(-10.437 + 2.280 ln(H) + 0.00532A - 0.000124A^2)
                    lln = np.exp( -10.437 + 2.280 * np.log(height) + 0.00532 * (age) - 0.000124 * (age ** 2) )
                case self.Parameters.FEV1_LLN:
                    # FEV1 L = exp(-9.111 + 2.014 ln(H) + 0.00102A - 0.000105A^2)
                    lln = np.exp(-9.111 + 2.014 * np.log(height) + 0.00102 * (age) - 0.000105 * (age ** 2) )
                case self.Parameters.MEF75_LLN:
                    # MEF75 L·s−1 = exp(-2.524 + 0.812 ln(H) + 0.00661A - 0.000132A^2)
                    lln = np.exp(- 2.524 + 0.812 * np.log(height) + 0.00661 * (age) - 0.000132 * (age ** 2) )
                case self.Parameters.MEF50_LLN:
                    # MEF50 L·s−1 = exp(-3.338 + 0.911 ln(H) - 0.00289A - 0.000109A^2)
                    lln = np.exp( -3.338 + 0.911 * np.log(height) - 0.00289 * (age) - 0.000109 * (age ** 2) )
                case self.Parameters.MEF25_LLN:
                    # MEF25 L·s−1 = exp(-4.262 + 1.009 ln(H) - 0.02485A - 0.000020A^2)
                    lln = np.exp( -4.262 + 1.009 * np.log(height) - 0.02485 * (age) -0.000020 * (age ** 2) )
                case self.Parameters.FEV1_FVC_LLN:
                    # FEV1/FVC % = exp(+6.180 - 0.341 ln(H) - 0.00529A + 0.000026A^
                    lln = np.exp( 6.180 - 0.341 * np.log(height) - 0.00529 * (age) + 0.000026 * (age ** 2) ) 
                case self.Parameters.PEF_LLN:
                    # PEF L·s−1 = exp(-3.992+1.170 ln(H)+0.00493A-0.000110A2)
                    lln = np.exp( -3.992 + 1.170 * np.log(height) + 0.00493 * (age) - 0.000110 * (age ** 2) )
                case _:
                    raise ValueError(f"Unknown parameter for percent calculation: {parameter}")
            return lln


import numpy
import pandas

from .src.KUSTER_2008 import KUSTER_2008
from .src.GLI_2012 import GLI_2012
from .src.SCHULZ_2013 import SCHULZ_2013
from .src.GLI_2017 import GLI_2017
from .src.GLI_2021 import GLI_2021
from .src.BOWERMANN_2022 import BOWERMANN_2022
from .src.SCAPIS_2023 import SCAPIS_2023


class Spiro:

    def __init__(self):
        print("""
This is the main object of pyspiro.
Please use a specific correction function instead of the main object.
As for example for GLI_2012:

import pandas
import GLI_2012 from pyspiro

gli = GLI_2012()

df = pandas.DataFrame(
    {"age": [2, 6, 7.15, 55, 60, 32.1], "sex": [1, 1, 1, 0, 0, 1], "height": [120, 160, 180, 130, 176, 160],
     "FEV1": [0.15, 1.241, 1.1, 0.8, 1.4, 1.2], "ethnicity": [1, 1, 1, 2, 3, 4],
     "FEF75": [0.15, 1.241, 1.1, 0.8, 1.4, 1.2]})

df["GLI_2012_FEV1"] = df.apply(
    lambda x: gli.percent(x.sex, x.age, x.height, 1, gli.Parameters["FEV1"], x.FEV1), axis=1)

df["GLI_2012_FEF75"] = df.apply(
    lambda x: gli.percent(x.sex, x.age, x.height, 1, gli.Parameters["FEF75"], x.FEF75), axis=1)

print(df)
        """)
        numpy.random.seed(42)
        n = 10  
        self.__dataframe = pandas.DataFrame({
            "age": numpy.random.randint(5, 80, size=n),
            "sex": numpy.random.choice([0, 1], size=n),
            "height": numpy.random.normal(170, 15, size=n).round(1),
            "VC": numpy.random.normal(3.5, 0.7, size=n).round(2),
            "RV": numpy.random.normal(1.5, 0.4, size=n).round(2),
            "TLC": numpy.random.normal(6.0, 0.9, size=n).round(2),
            "FEV1": numpy.random.normal(1.5, 3.0, size=n).round(2),
            "FEF75": numpy.random.normal(1.5, 3.0, size=n).round(2),
            "KCO": numpy.random.normal(1.5, 3.0, size=n).round(2),
            "ethnicity": numpy.random.choice([1, 4], size=n),
            "X10": numpy.random.normal(3.5, 0.7, size=n).round(2),
            "weight": numpy.random.normal(75,10, size = n).round(1),
        })

        self._kuster_2008_example()
        self._gli_2012_example()
        self._schulz_2013_example()
        self._gli_2017_example()
        self._gli_2021_example()
        self._bowermann_2022_example()
        self._scapis_2023_example()

        print(self.__dataframe)

    def _kuster_2008_example(self):

        kuster2008 = KUSTER_2008()
        kuster2008.set_silence(False)
        
        self.__dataframe["KUSTER_2008_FEV1"] = self.__dataframe.apply(
            lambda x: kuster2008.percent(x.sex, x.age, x.height, 1, kuster2008.Parameters.FEV1, x.FEV1), axis=1)

        self.__dataframe["KUSTER_2008_FEV1_LLN"] = self.__dataframe.apply(
            lambda x: kuster2008.lln(x.sex, x.age, x.height, 1, kuster2008.Parameters.FEV1_LLN, x.FEV1), axis=1)

    def _gli_2012_example(self):
        
        gli2012 = GLI_2012()
        gli2012.set_strategy("closest")

        self.__dataframe["GLI_2012_FEV1"] = self.__dataframe.apply(
            lambda x: gli2012.percent(x.sex, x.age, x.height, 1, gli2012.Parameters["FEV1"], x.FEV1), axis=1)

        self.__dataframe["GLI_2012_FEF75"] = self.__dataframe.apply(
            lambda x: gli2012.percent(x.sex, x.age, x.height, 1, gli2012.Parameters["FEF75"], x.FEF75), axis=1)

    def _gli_2017_example(self):
        
        gli2017 = GLI_2017()
        gli2017.set_strategy("closest")

        self.__dataframe["GLI_2017_KCO"] = self.__dataframe.apply(
            lambda x: gli2017.percent(x.sex, x.age, x.height, gli2017.Parameters.KCO_SI, x.KCO), axis=1)

    def _gli_2021_example(self):

        gli2021 = GLI_2021()

        self.__dataframe["GLI_2021_RV_p"] = self.__dataframe.apply(
            lambda x: gli2021.percent(x.sex, x.age, x.height, gli2021.Parameters.RV, x.RV), axis=1)

    def _bowermann_2022_example(self):
        
        bowermann = BOWERMANN_2022()

        self.__dataframe["BOWERMANN_FEV1_p"] = self.__dataframe.apply(
            lambda x: bowermann.percent(x.sex, x.age, x.height, bowermann.Parameters.FEV1, x.FEV1), axis=1)


    def _schulz_2013_example(self):

        schulz = SCHULZ_2013()

        self.__dataframe[["X10_05", "X10_50", "X10_95"]]  = self.__dataframe.apply(
            lambda x: pandas.Series(schulz.percentiles(x["sex"], x["age"], x["height"], x["weight"], schulz.Parameters.X10)), axis=1)

    def _scapis_2023_example(self):

        scapis = SCAPIS_2023()
        scapis.set_silence(False)
        scapis.set_strategy("closest")

        self.__dataframe["SCAPIS_FEV1_LLN"] = self.__dataframe.apply(
            lambda x: scapis.lln(x.sex, x.age, x.height, scapis.Parameters.pre_BD_FEV1, x.FEV1), axis=1)


if __name__ == '__main__':
    Spiro()


from .reference import Reference
from enum import Enum
import importlib.resources
import numpy
import pandas


class GLI_2017(Reference):
    """
    This GLI reference equations set was published in 2017 by Stanojevic and includes as variables sex, age, 
    height and ethnicity, providing functions for TLCO, DLCO, KCO and VA.

    It includes reference values for Caucasians aged 5-80 years.

    We here implement the corrected version of the reference values, which was published in 2020.

    Citation 1:
    Stanojevic S, Graham BL, Cooper BG, Thompson BR, Carter KW, Francis RW, Hall GL; Global Lung Function Initiative TLCO 
    working group; Global Lung Function Initiative (GLI) TLCO. Official ERS technical standards: Global Lung Function 
    Initiative reference values for the carbon monoxide transfer factor for Caucasians. Eur Respir J. 2017 
    Sep 11;50(3):1700010. doi: 10.1183/13993003.00010-2017. Erratum in: Eur Respir J. 2020 Oct 15;56(4):1750010. 
    doi: 10.1183/13993003.50010-2017. PMID: 28893868.

    Citation 2:
    "Official ERS technical standards: Global Lung Function Initiative reference values for the carbon monoxide transfer 
    factor for Caucasians." Sanja Stanojevic, Brian L. Graham, Brendan G. Cooper, Bruce R. Thompson, Kim W. Carter, Richard 
    W. Francis and Graham L. Hall on behalf of the Global Lung Function Initiative TLCO working group. Eur Respir J 2017; 
    50: 1700010. Eur Respir J. 2020 Oct 15;56(4):1750010. doi: 10.1183/13993003.50010-2017. Erratum for: Eur Respir J. 2017 
    Sep 11;50(3):1700010. doi: 10.1183/13993003.00010-2017. PMID: 33060163.
    """


    class Parameters(Enum):
        TLCO = 1 # SI Units
        DLCO = 2 # Traditional Units
        KCO_SI = 3 # SI Units
        KCO_trad = 4 # Traditional Units
        VA = 5

    def __init__(self):
        """
        No height range given.
        """
        self.__lookup, self.__splines = self.__load_lookup_table()

    def __load_lookup_table(self) -> tuple:
        """
        Loads and stores the coefficient and splines values.
        :return: both files as pandas dataframe
        """
        lookup_path = importlib.resources.open_binary('pyspiro.data', 'gli_2017_splines.csv')
        splines_path = importlib.resources.open_binary('pyspiro.data', 'gli_2017_coefficients.csv')
        lookup = pandas.read_csv(lookup_path, delimiter=";").set_index("age")
        splines = pandas.read_csv(splines_path, delimiter=";").set_index("var")
        self._age_range: tuple = (min(lookup.index), max(lookup.index))
        return lookup, splines

    def __get_splines(self, sex: int, age: float, parameter: int):
        """
        Yields the appropriated splines values based on the parameter, sex and age.
        :param sex: self.Sex enumration as integer value
        :param age: age as float
        :param parameter: self.Parameter enumeration as integer value
        """
        for i in ("Sspline", "Mspline", "Lspline"):
            yield self.__lookup["%s_%ss_%s" % (self.Parameters(parameter).name, self.Sex(sex).name.lower(), i)].loc[age] # Change sth here

    def lms(self, sex: int, age: float, height: float, parameter: int, value: float) -> tuple:
        """
        Calculate l, m and s values for the given parameters.
        """
        age = self.validate_range(round(age * 4) / 4, self._age_range, "age")
        if age is pandas.NA:
            return pandas.NA, pandas.NA, pandas.NA
        sspline, mspline, lspline = self.__get_splines(sex, age, parameter) #LSpline not used.
        c = self.__splines["%s_%ss" % (self.Parameters(parameter).name, self.Sex(sex).name.lower())]

        l = c.loc["q0"]
        m = numpy.exp(c.loc["a0"] + (c.loc["a1"] * numpy.log(height)) + (c.loc["a2"] * numpy.log(age)) + mspline)
        s = numpy.exp(c.loc["p0"] + (c.loc["p1"] * numpy.log(age)) + sspline)

        return l, m, s

    def percent(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns % of predicted value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else round(( value / m ) * 100, 2)

    def zscore(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns z-score value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else (((value/m)**l) - 1) / (l * s)

    def lln(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns lower limit of normal, by convention the lower 5th percentile.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))

    def all(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns all values at once (percent, z-score and lln).
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        if (l is pandas.NA or m is pandas.NA or s is pandas.NA):
            return pandas.NA
        else:
            return round(( value / m ) * 100, 2), (((value/m)**l) - 1) / (l * s), numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))


from setuptools import setup, find_packages
from pathlib import Path

VERSION = "0.3.6"
DESCRIPTION = "pyspiro"
this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / "README.md").read_text()

setup(
        name="pyspiro",
        version=VERSION,
        author="Hendrik Pott, Roman Martin",
        author_email="hendrik.pott@uni-marburg.de, roman.martin@uni-muenster.de",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        packages=find_packages(),
        package_data={"": ["data/test_file.csv", "data/pyspiro_250x.png",
        "data/gli_2012_splines.csv", "data/gli_2012_coefficients.csv",
        "data/schulz_2013_splines.csv",
        "data/gli_2017_splines.csv", "data/gli_2017_coefficients.csv",
        "data/gli_2021_splines.csv", "data/gli_2021_coefficients.csv",
        "data/bowermann_2022_splines.csv", "data/bowermann_2022_coefficients.csv",
        "data/scapis_2023_splines.csv", "data/scapis_2023_coefficients.csv"]},
        include_package_data=True,
        install_requires=["pandas", "numpy"],
        keywords=["python", "respirology", "spirometry", "bodyplethysmograph", "plethysmograph"],
        classifiers= [
            "Development Status :: 4 - Beta",
            "Intended Audience :: Healthcare Industry",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Education",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.10",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: Unix",
        ]
)



from abc import ABC, abstractmethod
from enum import Enum
from pandas import NA

class Reference(ABC):

    _strategy = "ignore"
    _silent = True

    def set_strategy(self, strategy: str):
        if strategy == "ignore" or strategy == "closest":
            self._strategy = strategy
            return True
        return False

    def set_silence(self, silent: bool):
        self._silent = silent

    def get_strategy(self):
        return self._strategy

    @abstractmethod
    def percent(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        pass

    @abstractmethod
    def zscore(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        pass

    @abstractmethod
    def lms(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        pass

    @abstractmethod
    def lln(self, sex: int, age: float, height: float, ethnicity: int, parameter: int, value: float):
        pass

    def check_range(self, value: float, value_range: tuple):
        return value_range[0] <= value <= value_range[1]

    def check_tuple(self, value: float, allowed: tuple, value_type: "value"):
        for i in allowed:
            if value == i:
                return value

        if not self._silent:
            print("The given %s of %.2f is not fitting to the allow values %s" % (value_type, value, str(allowed)))
        return NA

    def validate_range(self, value: float, value_range: tuple, value_type: str = "value"):
        if not self.check_range(value, value_range):
            if not self._silent:
                print("The given %s of %.2f does not fit to the defined %s range %.2f-%.2f" % (value_type, value, value_type, value_range[0], value_range[1]))

            if self._strategy == "closest":
                old_value = value
                if value <= value_range[0]:
                    value = value_range[0]
                else:
                    value = value_range[1]
                print("Set %s to %.2f from %.2f" % (value_type, value, old_value))
            elif self._strategy == "ignore":
                value = NA

        return value

    class Sex(Enum):
        FEMALE = 0
        MALE = 1


from .reference import Reference
from enum import Enum
import importlib.resources
import numpy
import pandas


class SCAPIS_2023(Reference):
    """
    This set of lung function equations was published in 2023 by Malinovschi et al and includes pre- and post-bronchodialtion reference equations for FEV1, FVC, FEV1_FVC, DLCO 
    and KCO for a population aged 50 to 65 (SCAPIS: Swedish CArdioPulmonary bioImage Study). Note that this set was derived from a Swedish population only and may not be applicable 
    to other populations.

    Citation:
    Malinovschi A, Zhou X, Andersson A, Backman H, Bake B, Blomberg A, Caidahl K, Eriksson MJ, Eriksson Ström J, Hamrefors V, Hjelmgren O, Janson C, Karimi R, Kylhammar D, Lindberg A, 
    Lindberg E, Liv P, Olin AC, Shalabi A, Sköld CM, Sundström J, Tanash H, Torén K, Wollmer P, Zaigham S, Östgren CJ, Engvall JE. Consequences of Using Post- or Prebronchodilator 
    Reference Values in Interpreting Spirometry. Am J Respir Crit Care Med. 2023 Aug 15;208(4):461-471. doi: 10.1164/rccm.202212-2341OC. PMID: 37339507.
    """

    class Parameters(Enum):
        """
        All parameters.
        """
        pre_BD_FEV1 = 1
        post_BD_FEV1 = 2
        pre_BD_FVC = 3
        post_BD_FVC = 4
        pre_BD_FEV1_FVC = 5
        post_BD_FEV1_FVC = 6
        post_BD_DLCO = 7
        post_BD_KCO = 18

    def __init__(self):
        """
        Initiate the object.
        """
        self.__lookup, self.__splines = self.__load_lookup_table()

    def __load_lookup_table(self) -> tuple:
        """
        Loads and stores the coefficient and splines values.
        :return: both files as pandas dataframe
        """
        lookup_path = importlib.resources.open_binary('pyspiro.data', 'scapis_2023_splines.csv')
        splines_path = importlib.resources.open_binary('pyspiro.data', 'scapis_2023_coefficients.csv')
        lookup = pandas.read_csv(lookup_path, delimiter=",", header = [0,1], index_col = 0)
        splines = pandas.read_csv(splines_path, delimiter=",", index_col=0)
        self._age_range: tuple = (min(lookup.index), max(lookup.index))
        return lookup, splines


    def __get_splines(self, sex: int, age: float, parameter: int):
        """
        Yields the appropriated splines values based on the parameter, sex and age.
        :param sex: self.Sex enumration as integer value
        :param age: age as float
        :param parameter: self.Parameter enumeration as integer value
        """
        for i in ("SSpline", "MSpline"):
            # No L-Splines
            yield self.__lookup.loc[age, ("%s_%s" % (self.Parameters(parameter).name, self.Sex(sex).name.lower()), i)]

    def lms(self, sex: int, age: float, height: float, parameter: int, value: float) -> tuple:
        """
        Calculate l, m and s values for the given parameters.
        """
        age = self.validate_range(round(age * 10) / 10, self._age_range, "age")
        if age is pandas.NA:
            return pandas.NA, pandas.NA, pandas.NA

        sspline, mspline = self.__get_splines(sex, age, parameter)
        c = self.__splines["%s_%s" % (self.Parameters(parameter).name, self.Sex(sex).name.lower())]
        
        m = numpy.exp(c.loc["M1"] + (c.loc["M2"] * numpy.log(height)) + (c.loc["M3"] * numpy.log(age)) + mspline)
        s = numpy.exp(c.loc["S1"] + (c.loc["S2"] * numpy.log(age)) + sspline)
        l = c.loc['L']
 
        return l, m, s

    def percent(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns % of predicted value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else round(( value / m ) * 100, 2)

    def zscore(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns z-score value.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else (((value/m)**l) - 1) / (l * s)

    def lln(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns lower limit of normal, by convention the lower 5th percentile.
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        return pandas.NA if (l is pandas.NA or m is pandas.NA or s is pandas.NA) else numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))

    def all(self, sex: int, age: float, height: float, parameter: int, value: float):
        """
        Returns all values at once (percent, z-score and lln).
        """
        l, m, s = self.lms(sex, age, height, parameter, value)
        if l is pandas.NA or m is pandas.NA or s is pandas.NA:
            return pandas.NA
        else:
            return round(( value / m ) * 100, 2), (((value/m)**l) - 1) / (l * s), numpy.exp(numpy.log(1 - 1.645 * l * s)/ l + numpy.log(m))

